/*
 * Copyright (c) 2013 Holger Weiss <holger@weiss.in-berlin.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef LIBUTIL_H
# define LIBUTIL_H

# if HAVE_CONFIG_H
#  include <config.h>
# endif

# include <sys/types.h>
# include <sys/stat.h>
# include <fcntl.h>

# ifndef HAVE_FLOCK
int flock(int, int);
#  ifndef LOCK_SH
#   define LOCK_SH 1
#  endif
#  ifndef LOCK_EX
#   define LOCK_EX 2
#  endif
#  ifndef LOCK_NB
#   define LOCK_NB 4
#  endif
#  ifndef LOCK_UN
#   define LOCK_UN 8
#  endif
# endif

struct pidfh;

int flopen(const char *, int, ...);
struct pidfh *pidfile_open(const char *, mode_t, pid_t *);
int pidfile_write(struct pidfh *);
int pidfile_remove(struct pidfh *);
int pidfile_close(struct pidfh *);
int pidfile_fileno(const struct pidfh *);

#endif

/* vim:set joinspaces noexpandtab textwidth=80 cinoptions=(4,u0: */
